"""Module for generating Fourier formulas.

This module provides the necessary routines for generating Fourier 
formulas; that is, `ndarrays` of `sympy.Expr` where each element
is a formula for a given Fourier component. The key benefit of working
with Fourier formulas rather than `phi`-space formulas is that all of
the fitting is easier -- instead of fitting a dataset to a complicated
trigonometric formula involving terms like ``cos(3*phi)``,
``sin**4(phi)``, etc., we simply fit the Fourier transform of the data
to the Fourier transform of the oritinal fitting formula. This results
not only in a drastic simplification of the fitting routines, but also
it turns out that computing the Fourier transform of a particular
formula is a problem that can be computed mostly ahead of time.

The way that this plays out in `shgpy` is that we compute the Fourier
transform of the uncontracted part of the response function -- that is,
the part of the response function which does not depend on the particular
form of the susceptbility tensor. This is conveniently the only
part which depends on `phi`. The result is a large (Fourier transformed)
tensor which we can save to a file on your machine using `pickle`. This
is done in :func:`~shgpy.fformgen.generate_uncontracted_fourier_transforms`
and :func:`~shgpy.fformgen.generate_uncontracted_fourier_transforms_symb`,
depending on whether the angle of incidence is hardcoded or a
`sympy.Symbol` instance.

Then, at runtime one can choose the susceptibility tensor they want to
fit (i.e. using :class:`~shgpy.tensor_definitions`) and simply contract
it (using :func:`~shgpy.fformgen.generate_contracted_fourier_transforms`
and :func:`~shgpy.fformgen.generate_contracted_fourier_transforms_complex`)
with the pickled formula to produce a Fourier formula. This can then
be manipulated and fitted at will (see :class:`~shgpy.fformfit`).

See the tutorial for more details.

"""
import sympy as sp
import numpy as np
from scipy.integrate import quad
from .core import (
    rotation_matrix3symb,
    n2i,
    tensor_contract,
    tensor_product,
)
from . import shg_symbols as S
import pickle
import sys
import logging
import time

_logger = logging.getLogger(__name__)


def _substitute_into_array(expr_array, *subs_tuples):
    ans = np.zeros(shape=expr_array.shape, dtype=object).flatten()
    temp = expr_array.flatten()
    for i in range(len(temp)):
        try:
            ans[i] = temp[i].subs(subs_tuples)
        except AttributeError:
            ans[i] = temp[i]
    return ans.reshape(expr_array.shape)


def _round_complex(z, ndigits):
    return round(np.real(z), ndigits)+1j*round(np.imag(z), ndigits)


def _round_expr(expr, ndigits):
    try:
        return expr.xreplace({n:(round(sp.re(n), ndigits)+1j*round(sp.im(n), ndigits)) for n in expr.atoms(sp.Number)})
    except AttributeError:
        return _round_complex(expr, ndigits)


def _round_complex_tensor(t, ndigits):
    ans = t.flatten()
    for i in range(len(ans)):
        try:
            ans[i] = _round_expr(ans[i], ndigits)
        except AttributeError:
            ans[i] = round(sp.re(ans[i]), ndigits)+1j*round(sp.im(ans[i]), ndigits)
    ans = ans.reshape(t.shape)
    return ans


def _conjugate_tensor(tensor):
    ans = np.zeros(len(tensor.flatten()), dtype=object)
    for i,expr in enumerate(tensor.flatten()):
        ans[i] = sp.conjugate(sp.sympify(expr))
    return ans.reshape(tensor.shape)


def _free_symbols_of_array(array):
    total = []
    for a in array.flatten():
        total = total+list(sp.sympify(a).free_symbols)
    return set(total)


def _fexpr_n(expr_arr, n, precision=7):
    Rshape = expr_arr.shape
    expr_arrf = expr_arr.flatten()
    h = np.zeros(len(expr_arrf), dtype=object)
    for i in range(len(expr_arrf)):
        f_re = sp.lambdify(S.phi, 1/2/sp.pi*expr_arrf[i]*sp.cos(-1*(n*S.phi)))
        f_im = sp.lambdify(S.phi, 1/2/sp.pi*expr_arrf[i]*sp.sin(-1*(n*S.phi)))
        t_re,_ = quad(f_re, 0, 2*np.pi)
        t_im,_ = quad(f_im, 0, 2*np.pi)
        h[i] = round(t_re, precision)+round(t_im, precision)*1j
    h = h.reshape(Rshape)
    return h


def _convolve_ftensors(nR1, nR2, M=16, dtype=object):
    test_prod = tensor_product(nR1[0], nR2[0])
    ans = np.zeros(dtype=dtype, shape=(2*M+1,)+test_prod.shape)
    for n in np.arange(-M, M+1):
        for m in np.arange(-M, M+1):
            try:
                ans[n2i(n, M)] += tensor_product(nR1[n2i(m, M)], nR2[n2i(n-m, M)])
            except IndexError:
                pass
    return ans


def _load_pickle(filename):
    return np.load(filename, allow_pickle=True)


def _save_fform_dict(filename, _fform_dict):
    pickle.dump(_fform_dict, open(filename, 'wb'))


def _load_fform_dict(filename):
    return pickle.load(open(filename, 'rb'))


def generate_uncontracted_fourier_transforms(aoi, uncontracted_filename_prefix, M=16):
    """Generate uncontracted Fourier transforms.

    Parameters
    ----------
    aoi : float
        Angle of incidence (in radians). This is hardcoded in this
        function but can be left as a free parameter in :func:`~shgpy.fformgen.generate_uncontracted_fourier_transforms_symb`.
    uncontracted_filename_prefix : str
        Filename prefix for where to save the pickled Fourier
        transforms. The pickle objects are saved to
        ``uncontracted_filename_prefix+'_pp'``, ``uncontracted_filename_prefix+'_ps'``, etc.
    M : int, optional
        Number of Fourier frequencies to compute. Default is 16.
        The Fourier formulas produced with this function have ``2*M+1``
        elements each.

    Notes
    -----
    Saves uncontracted Fourier transforms using `pickle` to the
    specified filenames.

    """
    include_quadrupole = True
    start = time.time()

    ## 
    ## First define all of the types of 
    ## tensors we will need to form the
    ## tensor product.
    ## 
    F = np.array([S.Fx, S.Fy, S.Fz])
    np.set_printoptions(threshold=sys.maxsize)
    R = np.array(rotation_matrix3symb([0, 0, 1], S.phi, ndigits=5))
    Id = np.identity(3)
    k_out = _substitute_into_array(np.array([-sp.sin(S.theta), 0, -sp.cos(S.theta)]), (S.theta, aoi))
    k_in = _substitute_into_array(np.array([-sp.sin(S.theta), 0, sp.cos(S.theta)]), (S.theta, aoi))
    proj = Id - tensor_product(k_out, k_out)
    proj_x = proj[0]
    proj_y = proj[1]
    proj_z = proj[2]

    ##
    ## Now take the fourier transform
    ## of all the tensors above. For
    ## tensors which are independent of
    ## phi, of course the fourier transform
    ## just has an m=0 component.
    ##
    rproj = np.zeros(shape=(2*M+1,)+proj.shape, dtype=object)
    rproj[n2i(0, M)] = proj
    rproj_x = np.zeros(shape=(2*M+1,)+proj_x.shape, dtype=object)
    rproj_x[n2i(0, M)] = proj_x
    rproj_y = np.zeros(shape=(2*M+1,)+proj_y.shape, dtype=object)
    rproj_y[n2i(0, M)] = proj_y
    rproj_z = np.zeros(shape=(2*M+1,)+proj_z.shape, dtype=object)
    rproj_z[n2i(0, M)] = proj_z
    rR = [_fexpr_n(R, m) for m in np.arange(-M, M+1)]
    rF = np.zeros(shape=(2*M+1,)+F.shape, dtype=object)
    rF[n2i(0, M)] = F
    rFc = np.zeros(shape=(2*M+1,)+F.shape, dtype=object)
    rFc[n2i(0, M)] = sp.conjugate(F)
    r1_x = _convolve_ftensors(rproj_x, rR)
    r1_y = _convolve_ftensors(rproj_y, rR)
    r1_z = _convolve_ftensors(rproj_z, rR)
    rk_out = np.zeros(shape=(2*M+1,)+k_out.shape, dtype=object)
    rk_out[n2i(0, M)] = k_out
    rk_in = np.zeros(shape=(2*M+1,)+k_in.shape, dtype=object)
    rk_in[n2i(0, M)] = k_in

    ## 
    ## This is so I can address the P and
    ## S components individually.
    ## 
    r1_0 = np.zeros(shape=r1_x.shape, dtype=object)
    r1_s = np.array([np.array([r1_0[n2i(m, M)], r1_y[n2i(m, M)], r1_0[n2i(m, M)]]) for m in np.arange(-M, M+1)])
    r1_p = np.array([np.array([r1_x[n2i(m, M)], r1_0[n2i(m, M)], r1_z[n2i(m, M)]]) for m in np.arange(-M, M+1)])

    ##
    ## Now I am ready to do the long tensor
    ## contraction. I have found that the
    ## computation is much faster if you
    ## resort to small tensor products and build
    ## up the long contraction piece by piece.
    ## In the future I would like to build
    ## a method to do all of this with one
    ## function call, but for now this will do...
    ##
    h7_arr_term1 = []
    h7_arr_term2 = []
    h7_arr_term3 = []
    h7_arr_term4 = []
    for r1 in [r1_p, r1_s]:
        h1 = np.array([tensor_contract(r1[n2i(m, M)], [[1, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h1 done.')
        r2 = _convolve_ftensors(h1, h1)
        h2 = np.array([tensor_contract(r2[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h2 done.')
        r3 = _convolve_ftensors(rR, rF)
        h3 = np.array([tensor_contract(r3[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h3 done.')
        h4 = _convolve_ftensors(h2, h3)
        _logger.debug('h4 done.')
        h5 = _convolve_ftensors(h4, h3)
        _logger.debug('h5 done.')
        h6 = _convolve_ftensors(h5, h3)
        _logger.debug('h6 done.')
        h7 = _convolve_ftensors(h6, h3)
        _logger.debug('h7 done.')
        h7_arr_term1.append(h7)
        if include_quadrupole is True:
            r4 = _convolve_ftensors(rR, -1j*rk_in)
            h8 = np.array([tensor_contract(r4[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
            _logger.debug('h8 done.')
            h9 = _convolve_ftensors(h7, h8)
            _logger.debug('h9 done.')
            h7_arr_term2.append(h9)
            h7_arr_term3.append(-1*h9)
            h10 = _convolve_ftensors(h9, -1*h8)
            _logger.debug('h10 done.')
            h7_arr_term4.append(h10)
        
    ##
    ## Now I form four of these tensors, each
    ## pertaining to a particular polarization
    ## combination.
    ##
    list_of_terms = [h7_arr_term1, h7_arr_term2, h7_arr_term3, h7_arr_term4]
    nterms = len(list_of_terms)
    _logger.debug('Started substitution.')

    h7_pp = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_pp[i] = _substitute_into_array(h7_arr[0], (S.Fx, np.cos(aoi)), (S.Fy, 0), (S.Fz, np.sin(aoi)))
    _logger.debug('done 1.')

    h7_ps = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_ps[i] = _substitute_into_array(h7_arr[1], (S.Fx, np.cos(aoi)), (S.Fy, 0), (S.Fz, np.sin(aoi)))
    _logger.debug('done 2.')
    
    h7_sp = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_sp[i] = _substitute_into_array(h7_arr[0], (S.Fx, 0), (S.Fy, -1), (S.Fz, 0))
    _logger.debug('done 3.')

    h7_ss = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_ss[i] = _substitute_into_array(h7_arr[1], (S.Fx, 0), (S.Fy, -1), (S.Fz, 0))
    _logger.debug('done 4.')

    _logger.debug(f'Generation of uncontracted fourier transforms completed. It took {time.time()-start} seconds.')
    
    ##
    ## Now to generate the fourier transformed
    ## SHG formulae for a particular tensor chi,
    ## all you need to do is to contract these
    ## tensors in a certain way with chi x chi.
    ##
    ## All that's left is to save these h7 tensors
    ## to a file.
    ##
    np.save(uncontracted_filename_prefix+'_pp', h7_pp)
    np.save(uncontracted_filename_prefix+'_ps', h7_ps)
    np.save(uncontracted_filename_prefix+'_sp', h7_sp)
    np.save(uncontracted_filename_prefix+'_ss', h7_ss)


def generate_uncontracted_fourier_transforms_symb(uncontracted_filename_prefix, M=16):
    """Generate uncontracted Fourier transforms with symbolic angle of incidence.

    Parameters
    ----------
    uncontracted_filename_prefix : str
        Filename prefix for where to save the pickled Fourier
        transforms. The pickle objects are saved to
        ``uncontracted_filename_prefix+'_pp'``, ``uncontracted_filename_prefix+'_ps'``, etc.
    M : int, optional
        Number of Fourier frequencies to compute. Default is 16.
        The Fourier formulas produced with this function have ``2*M+1``
        elements each.

    Notes
    -----
    Saves uncontracted Fourier transforms using `pickle` to the
    specified filenames. The angle of incidence is hardcoded to
    be the `sympy.Symbol` `theta` defined in :class:`~shgpy.shg_symbols`.

    """
    start = time.time()
    include_quadrupole = True

    ## 
    ## First define all of the types of 
    ## tensors we will need to form the
    ## tensor product.
    ## 
    F = np.array([S.Fx, S.Fy, S.Fz])
    np.set_printoptions(threshold=sys.maxsize)
    R = np.array(rotation_matrix3symb([0, 0, 1], S.phi, ndigits=5))
    Id = np.identity(3)
##     k_out = _substitute_into_array(np.array([-sp.sin(theta), 0, -sp.cos(theta)]), (theta, aoi))
##     k_in = _substitute_into_array(np.array([-sp.sin(theta), 0, sp.cos(theta)]), (theta, aoi))
    k_out = np.array([-sp.sin(S.theta), 0, -sp.cos(S.theta)])
    k_in = np.array([-sp.sin(S.theta), 0, sp.cos(S.theta)])
    proj = Id - tensor_product(k_out, k_out)
    proj_x = proj[0]
    proj_y = proj[1]
    proj_z = proj[2]

    ##
    ## Now take the fourier transform
    ## of all the tensors above. For
    ## tensors which are independent of
    ## phi, of course the fourier transform
    ## just has an m=0 component.
    ##
    rproj = np.zeros(shape=(2*M+1,)+proj.shape, dtype=object)
    rproj[n2i(0, M)] = proj
    rproj_x = np.zeros(shape=(2*M+1,)+proj_x.shape, dtype=object)
    rproj_x[n2i(0, M)] = proj_x
    rproj_y = np.zeros(shape=(2*M+1,)+proj_y.shape, dtype=object)
    rproj_y[n2i(0, M)] = proj_y
    rproj_z = np.zeros(shape=(2*M+1,)+proj_z.shape, dtype=object)
    rproj_z[n2i(0, M)] = proj_z
    rR = [_fexpr_n(R, m) for m in np.arange(-M, M+1)]
    rF = np.zeros(shape=(2*M+1,)+F.shape, dtype=object)
    rF[n2i(0, M)] = F
    rFc = np.zeros(shape=(2*M+1,)+F.shape, dtype=object)
    rFc[n2i(0, M)] = sp.conjugate(F)
    r1_x = _convolve_ftensors(rproj_x, rR)
    r1_y = _convolve_ftensors(rproj_y, rR)
    r1_z = _convolve_ftensors(rproj_z, rR)
    rk_out = np.zeros(shape=(2*M+1,)+k_out.shape, dtype=object)
    rk_out[n2i(0, M)] = k_out
    rk_in = np.zeros(shape=(2*M+1,)+k_in.shape, dtype=object)
    rk_in[n2i(0, M)] = k_in

    ## 
    ## This is so I can address the P and
    ## S components individually.
    ## 
    r1_0 = np.zeros(shape=r1_x.shape, dtype=object)
    r1_s = np.array([np.array([r1_0[n2i(m, M)], r1_y[n2i(m, M)], r1_0[n2i(m, M)]]) for m in np.arange(-M, M+1)])
    r1_p = np.array([np.array([r1_x[n2i(m, M)], r1_0[n2i(m, M)], r1_z[n2i(m, M)]]) for m in np.arange(-M, M+1)])

    ##
    ## Now I am ready to do the long tensor
    ## contraction. I have found that the
    ## computation is much faster if you
    ## resort to small tensor products and build
    ## up the long contraction piece by piece.
    ## In the future I would like to build
    ## a method to do all of this with one
    ## function call, but for now this will do...
    ##
    h7_arr_term1 = []
    h7_arr_term2 = []
    h7_arr_term3 = []
    h7_arr_term4 = []
    for r1 in [r1_p, r1_s]:
        h1 = np.array([tensor_contract(r1[n2i(m, M)], [[1, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h1 done.')
        r2 = _convolve_ftensors(h1, h1)
        h2 = np.array([tensor_contract(r2[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h2 done.')
        r3 = _convolve_ftensors(rR, rF)
        h3 = np.array([tensor_contract(r3[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
        _logger.debug('h3 done.')
        h4 = _convolve_ftensors(h2, h3)
        _logger.debug('h4 done.')
        h5 = _convolve_ftensors(h4, h3)
        _logger.debug('h5 done.')
        h6 = _convolve_ftensors(h5, h3)
        _logger.debug('h6 done.')
        h7 = _convolve_ftensors(h6, h3)
        _logger.debug('h7 done.')
        h7_arr_term1.append(h7)
        if include_quadrupole is True:
            r4 = _convolve_ftensors(rR, -1j*rk_in)
            h8 = np.array([tensor_contract(r4[n2i(m, M)], [[0, 2]]) for m in np.arange(-M, M+1)])
            _logger.debug('h8 done.')
            h9 = _convolve_ftensors(h7, h8)
            _logger.debug('h9 done.')
            h7_arr_term2.append(h9)
            h7_arr_term3.append(-1*h9)
            h10 = _convolve_ftensors(h9, -1*h8)
            _logger.debug('h10 done.')
            h7_arr_term4.append(h10)
        
    ##
    ## Now I form four of these tensors, each
    ## pertaining to a particular polarization
    ## combination.
    ##
    list_of_terms = [h7_arr_term1, h7_arr_term2, h7_arr_term3, h7_arr_term4]
    nterms = len(list_of_terms)
    _logger.debug('Started substitution.')

    h7_pp = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_pp[i] = _substitute_into_array(h7_arr[0], (S.Fx, sp.cos(S.theta)), (S.Fy, 0), (S.Fz, sp.sin(S.theta)))
    _logger.debug('done 1.')

    h7_ps = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_ps[i] = _substitute_into_array(h7_arr[1], (S.Fx, sp.cos(S.theta)), (S.Fy, 0), (S.Fz, sp.sin(S.theta)))
    _logger.debug('done 2.')
    
    h7_sp = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_sp[i] = _substitute_into_array(h7_arr[0], (S.Fx, 0), (S.Fy, -1), (S.Fz, 0))
    _logger.debug('done 3.')

    h7_ss = np.empty(nterms, dtype=object)
    for i,h7_arr in enumerate(list_of_terms):
        h7_ss[i] = _substitute_into_array(h7_arr[1], (S.Fx, 0), (S.Fy, -1), (S.Fz, 0))
    _logger.debug('done 4.')

    _logger.debug(f'Generation of uncontracted fourier transforms completed. It took {time.time()-start} seconds.')
    
    ##
    ## Now to generate the fourier transformed
    ## SHG formulae for a particular tensor chi,
    ## all you need to do is to contract these
    ## tensors in a certain way with chi x chi.
    ##
    ## All that's left is to save these h7 tensors
    ## to a file.
    ##
    np.save(uncontracted_filename_prefix+'_pp', h7_pp)
    np.save(uncontracted_filename_prefix+'_ps', h7_ps)
    np.save(uncontracted_filename_prefix+'_sp', h7_sp)
    np.save(uncontracted_filename_prefix+'_ss', h7_ss)


def generate_contracted_fourier_transforms(save_filename, uncontracted_filename_prefix, chi_dipole, chi_quadrupole, M=16, ndigits=None):
    """Generate contracted Fourier transforms.

    Parameters
    ----------
    save_filename : str or file object
        Filename to save the pickled Fourier formula to.
    uncontracted_filename_prefix : str
        The filename prefix of the pickled uncontracted Fourier
        transforms generated by :func:`~shgpy.fformgen.generate_uncontracted_fourier_transforms`.    
    chi_dipole : ndarray
        Dipole susceptibility tensor
    chi_quadupole : ndarray
        Quadrupole susceptibility tensor
    M : int, optional
        Number of Fourier frequencies to compute. Default is 16.
        The Fourier formulas produced with this function have ``2*M+1``
        elements each.
    ndigits : int, optional
        Number of digits to round to for each term in the Fourier formula.
        If `None`, no rounding is done. Default is `None`.

    Notes
    -----
    Saves contracted Fourier transforms using `pickle` to `save_filename`. 

    """
    ##
    ## Now I build a useful set of arrays
    ## which we will use later to reduce
    ## the number of lines in this program
    ##
    pcs = ['PP', 'PS', 'SP', 'SS']
    terms_dict = {pc:_load_pickle(uncontracted_filename_prefix+'_'+pc.lower()+'.npy') for pc in pcs}

    contraction_lists_1 = [[[0, 6], [2, 7], [3, 8]],
                           [[0, 7], [2, 8], [3, 9]],
                           [[0, 7], [2, 9], [3, 10], [6, 8]],
                           [[0, 8], [2, 10], [3, 11], [6, 9]]]
    contraction_lists_2 = [[[0, 3], [1, 4], [2, 5]],
                           [[0, 4], [1, 6], [2, 7], [3, 5]],
                           [[0, 3], [1, 4], [2, 5]],
                           [[0, 4], [1, 6], [2, 7], [3, 5]]]

    chi_list_1 = [chi_dipole, chi_dipole, chi_quadrupole, chi_quadrupole]
    chi_list_2 = [chi_dipole, chi_quadrupole, chi_dipole, chi_quadrupole]

    _logger.debug('Finished preparation.')

    ##
    ## Now we do the contraction with chi x chi,
    ## for each polarization combination and for
    ## each term. Each of the tensors h7_pp, ...
    ## has 4 elements, correponsing to the four
    ## terms in |P+ikQ|^2.
    ##
    _fform_dict = {}
    for pc,h7_pc in terms_dict.items():
        _fform_dict[pc] = np.zeros(shape=(2*M+1,), dtype=object)
        for term in range(len(h7_pc)):
            t8_pc_term = np.array([tensor_contract(tensor_product(h7_pc[term][n2i(m, M)], chi_list_1[term]), contraction_lists_1[term]) for m in np.arange(-M, M+1)])
            t9_pc_term = np.array([tensor_contract(tensor_product(t8_pc_term[n2i(m, M)], chi_list_2[term]), contraction_lists_2[term]) for m in np.arange(-M, M+1)])
            _fform_dict[pc] += np.copy(t9_pc_term)
            _logger.debug('Finished term %s.' % term)
        if ndigits is not None:
            _fform_dict[pc] = _round_complex_tensor(_fform_dict[pc], ndigits)
        _logger.debug('Finished %s.' % pc)

    _save_fform_dict(save_filename, _fform_dict)


def generate_contracted_fourier_transforms_complex(save_filename, uncontracted_filename_prefix, chi_dipole, chi_quadrupole, M=16, ndigits=None):
    """Generate contracted Fourier transforms assuming complex coefficients.

    Parameters
    ----------
    save_filename : str or file object
        Filename to save the pickled Fourier formula to.
    uncontracted_filename_prefix : str
        The filename prefix of the pickled uncontracted Fourier
        transforms generated by :func:`~shgpy.fformgen.generate_uncontracted_fourier_transforms`.    
    chi_dipole : ndarray
        Dipole susceptibility tensor
    chi_quadupole : ndarray
        Quadrupole susceptibility tensor
    M : int, optional
        Number of Fourier frequencies to compute. Default is 16.
        The Fourier formulas produced with this function have ``2*M+1``
        elements each.
    ndigits : int, optional
        Number of digits to round to for each term in the Fourier formula.
        If `None`, no rounding is done. Default is `None`.

    Raises
    ------
    ValueError
        If not all parameters of `chi_dipole` and `chi_quadrupole`
        hardcoded as real, i.e. you must specifically indicate the
        real and imaginary parts of each variable in `chi_dipole` and
        `chi_quadrupole` (see :func:`~shgpy.core.utilities.make_tensor_complex`).

    Notes
    -----
    Duplicate of :func:`~shgpy.fformgen.generate_contracted_fourier_transforms`
    with complex coefficients. See note in :func:`~shgpy.formgen.formgen_just_dipole_complex`.
    Saves contracted Fourier transforms using `pickle` to `save_filename`. 

    """
    ##
    ## First we check if all the parameters in 
    ## chi_dipole and chi_quadrupole are real.
    ## If not, raise an error.
    ##
    for chi in [chi_dipole, chi_quadrupole]:
        free_symbols = _free_symbols_of_array(chi)
        for fs in free_symbols:
            if fs.is_real is not True:
                raise ValueError('Parameters of chi must all be real: %s' % str(fs))

    ##
    ## Now I build a useful set of arrays
    ## which we will use later to reduce
    ## the number of lines in this program
    ##
    pcs = ['PP', 'PS', 'SP', 'SS']
    terms_dict = {pc:_load_pickle(uncontracted_filename_prefix+'_'+pc.lower()+'.npy') for pc in pcs}

    contraction_lists_1 = [[[0, 6], [2, 7], [3, 8]],
                           [[0, 7], [2, 8], [3, 9]],
                           [[0, 7], [2, 9], [3, 10], [6, 8]],
                           [[0, 8], [2, 10], [3, 11], [6, 9]]]
    contraction_lists_2 = [[[0, 3], [1, 4], [2, 5]],
                           [[0, 4], [1, 6], [2, 7], [3, 5]],
                           [[0, 3], [1, 4], [2, 5]],
                           [[0, 4], [1, 6], [2, 7], [3, 5]]]

    chi_list_1 = [chi_dipole, chi_dipole, chi_quadrupole, chi_quadrupole]
    chi_list_2 = [_conjugate_tensor(chi_dipole), _conjugate_tensor(chi_quadrupole), _conjugate_tensor(chi_dipole), _conjugate_tensor(chi_quadrupole)]

    _logger.debug('Finished preparation.')

    ##
    ## Now we do the contraction with chi x chi,
    ## for each polarization combination and for
    ## each term. Each of the tensors h7_pp, ...
    ## has 4 elements, correponsing to the four
    ## terms in |P+ikQ|^2.
    ##
    _fform_dict = {}
    for pc,h7_pc in terms_dict.items():
        _fform_dict[pc] = np.zeros(shape=(2*M+1,), dtype=object)
        for term in range(len(h7_pc)):
            t8_pc_term = np.array([tensor_contract(tensor_product(h7_pc[term][n2i(m, M)], chi_list_1[term]), contraction_lists_1[term]) for m in np.arange(-M, M+1)])
            t9_pc_term = np.array([tensor_contract(tensor_product(t8_pc_term[n2i(m, M)], chi_list_2[term]), contraction_lists_2[term]) for m in np.arange(-M, M+1)])
            _fform_dict[pc] += np.copy(t9_pc_term)
            _logger.debug('Finished term %s.' % term)
        if ndigits is not None:
            _fform_dict[pc] = _round_complex_tensor(_fform_dict[pc], ndigits)
        _logger.debug('Finished %s.' % pc)

    _save_fform_dict(save_filename, _fform_dict)



import unittest

import random
import logging
import numpy as np

import shgpy
import shgpy.tensor_definitions as td
from shgpy.formgen import (
    formgen_just_dipole_complex,
    formgen_just_dipole_real,
    formgen_dipole_quadrupole_complex,
    formgen_dipole_quadrupole_real,
)
import shgpy.shg_symbols as S
from shgpy.plotter import easy_plot
import shgpy.fformfit

show_plot = False

logging.getLogger(__name__)

MANUAL = False


class TestFormGen(unittest.TestCase):

    def test_formgen(self):
        t_dip = td.dipole['T_d']
        t_quad = td.quadrupole['T_d']
        form1 = formgen_just_dipole_real(t_dip, S.theta)
        form2 = formgen_dipole_quadrupole_real(t_dip, t_quad, S.theta)
        
        t_dip = shgpy.make_tensor_complex(td.dipole['T_d'])
        t_quad = shgpy.make_tensor_complex(td.quadrupole['T_d'])
        form3 = formgen_just_dipole_complex(t_dip, S.theta)
        form4 = formgen_dipole_quadrupole_complex(t_dip, t_quad, S.theta)
        forms = [form1, form2, form3, form4]
        for form in forms:
            form.apply_phase_shift(S.psi, S.phi)
            dat = shgpy.form_to_dat(form, {k:random.uniform(-1, 1) for k in form.get_free_symbols() if k != S.phi}, 1000)
            if MANUAL:
                easy_plot([dat], [{'linestyle':'-', 'color':'blue'}], dat.get_keys(), show_plot=False)

    def test_formfit(self):
        R = shgpy.rotation_matrix_from_two_vectors(np.array([1, 1, 0]), np.array([0, 0, 1]))
        t_dip = shgpy.transform(shgpy.particularize(td.dipole['T_d']), R)
        form = formgen_just_dipole_real(t_dip, 0.1745)
        fform = shgpy.form_to_fform(form)
        fform.apply_phase_shift(S.psi)
        data_filenames_dict = {
            'PP':'tests/Data/dataPP.csv',
            'PS':'tests/Data/dataPS.csv',
            'SP':'tests/Data/dataSP.csv',
            'SS':'tests/Data/dataSS.csv',
        }
        dat, fdat = shgpy.load_data_and_fourier_transform(data_filenames_dict, 'degrees')
        guess_dict = {k:1 for k in fform.get_free_symbols()}
        ret = shgpy.fformfit.least_squares_fit(fform, fdat, guess_dict)
        self.assertAlmostEqual(abs(ret.xdict[S.psi]), 1.59, delta=0.1)
        self.assertAlmostEqual(abs(ret.xdict[S.zyx]), 1.23, delta=0.1)
        fit_dat = shgpy.fform_to_dat(fform, ret.xdict, 1000)
        if MANUAL:
            easy_plot([dat, fit_dat], [{'linestyle':' ', 'markerfacecolor':'none', 'color':'blue', 'marker':'o'}, {'linestyle':'-', 'color':'blue'}], ['PP', 'PS', 'SP', 'SS'], show_plot=True, filename=None, show_legend=False)



"""This module defines a number of sympy.Symbol objects
which are used throughout `shgpy`. They are used to define
free parameters in suscepitbility tensors (see :class:`~shgpy.tensor_definitions`),
and to standardize name conventions for important variables
(like `phi` and `theta`, which are used throughout shgpy as the standard
azimuthal and incidence angles, respectively).

These variables in this module should be treated as protected and not
modified by the user unless absolutely necessary.

"""
import sympy as sp

xxx = sp.symbols('xxx')
""""""
xxy = sp.symbols('xxy')
""""""
xxz = sp.symbols('xxz')
""""""
xyx = sp.symbols('xyx')
""""""
xyy = sp.symbols('xyy')
""""""
xyz = sp.symbols('xyz')
""""""
xzx = sp.symbols('xzx')
""""""
xzy = sp.symbols('xzy')
""""""
xzz = sp.symbols('xzz')
""""""
yxx = sp.symbols('yxx')
""""""
yxy = sp.symbols('yxy')
""""""
yxz = sp.symbols('yxz')
""""""
yyx = sp.symbols('yyx')
""""""
yyy = sp.symbols('yyy')
""""""
yyz = sp.symbols('yyz')
""""""
yzx = sp.symbols('yzx')
""""""
yzy = sp.symbols('yzy')
""""""
yzz = sp.symbols('yzz')
""""""
zxx = sp.symbols('zxx')
""""""
zxy = sp.symbols('zxy')
""""""
zxz = sp.symbols('zxz')
""""""
zyx = sp.symbols('zyx')
""""""
zyy = sp.symbols('zyy')
""""""
zyz = sp.symbols('zyz')
""""""
zzx = sp.symbols('zzx')
""""""
zzy = sp.symbols('zzy')
""""""
zzz = sp.symbols('zzz')
""""""

sxxx = sp.symbols('sxxx')
""""""
sxxy = sp.symbols('sxxy')
""""""
sxxz = sp.symbols('sxxz')
""""""
sxyx = sp.symbols('sxyx')
""""""
sxyy = sp.symbols('sxyy')
""""""
sxyz = sp.symbols('sxyz')
""""""
sxzx = sp.symbols('sxzx')
""""""
sxzy = sp.symbols('sxzy')
""""""
sxzz = sp.symbols('sxzz')
""""""
syxx = sp.symbols('syxx')
""""""
syxy = sp.symbols('syxy')
""""""
syxz = sp.symbols('syxz')
""""""
syyx = sp.symbols('syyx')
""""""
syyy = sp.symbols('syyy')
""""""
syyz = sp.symbols('syyz')
""""""
syzx = sp.symbols('syzx')
""""""
syzy = sp.symbols('syzy')
""""""
syzz = sp.symbols('syzz')
""""""
szxx = sp.symbols('szxx')
""""""
szxy = sp.symbols('szxy')
""""""
szxz = sp.symbols('szxz')
""""""
szyx = sp.symbols('szyx')
""""""
szyy = sp.symbols('szyy')
""""""
szyz = sp.symbols('szyz')
""""""
szzx = sp.symbols('szzx')
""""""
szzy = sp.symbols('szzy')
""""""
szzz = sp.symbols('szzz')
""""""

xxxx= sp.symbols('xxxx')
""""""
xxxy= sp.symbols('xxxy')
""""""
xxxz= sp.symbols('xxxz')
""""""
xxyx= sp.symbols('xxyx')
""""""
xxyy= sp.symbols('xxyy')
""""""
xxyz= sp.symbols('xxyz')
""""""
xxzx= sp.symbols('xxzx')
""""""
xxzy= sp.symbols('xxzy')
""""""
xxzz= sp.symbols('xxzz')
""""""
xyxx= sp.symbols('xyxx')
""""""
xyxy= sp.symbols('xyxy')
""""""
xyxz= sp.symbols('xyxz')
""""""
xyyx= sp.symbols('xyyx')
""""""
xyyy= sp.symbols('xyyy')
""""""
xyyz= sp.symbols('xyyz')
""""""
xyzx= sp.symbols('xyzx')
""""""
xyzy= sp.symbols('xyzy')
""""""
xyzz= sp.symbols('xyzz')
""""""
xzxx= sp.symbols('xzxx')
""""""
xzxy= sp.symbols('xzxy')
""""""
xzxz= sp.symbols('xzxz')
""""""
xzyx= sp.symbols('xzyx')
""""""
xzyy= sp.symbols('xzyy')
""""""
xzyz= sp.symbols('xzyz')
""""""
xzzx= sp.symbols('xzzx')
""""""
xzzy= sp.symbols('xzzy')
""""""
xzzz= sp.symbols('xzzz')
""""""
yxxx= sp.symbols('yxxx')
""""""
yxxy= sp.symbols('yxxy')
""""""
yxxz= sp.symbols('yxxz')
""""""
yxyx= sp.symbols('yxyx')
""""""
yxyy= sp.symbols('yxyy')
""""""
yxyz= sp.symbols('yxyz')
""""""
yxzx= sp.symbols('yxzx')
""""""
yxzy= sp.symbols('yxzy')
""""""
yxzz= sp.symbols('yxzz')
""""""
yyxx= sp.symbols('yyxx')
""""""
yyxy= sp.symbols('yyxy')
""""""
yyxz= sp.symbols('yyxz')
""""""
yyyx= sp.symbols('yyyx')
""""""
yyyy= sp.symbols('yyyy')
""""""
yyyz= sp.symbols('yyyz')
""""""
yyzx= sp.symbols('yyzx')
""""""
yyzy= sp.symbols('yyzy')
""""""
yyzz= sp.symbols('yyzz')
""""""
yzxx= sp.symbols('yzxx')
""""""
yzxy= sp.symbols('yzxy')
""""""
yzxz= sp.symbols('yzxz')
""""""
yzyx= sp.symbols('yzyx')
""""""
yzyy= sp.symbols('yzyy')
""""""
yzyz= sp.symbols('yzyz')
""""""
yzzx= sp.symbols('yzzx')
""""""
yzzy= sp.symbols('yzzy')
""""""
yzzz= sp.symbols('yzzz')
""""""
zxxx= sp.symbols('zxxx')
""""""
zxxy= sp.symbols('zxxy')
""""""
zxxz= sp.symbols('zxxz')
""""""
zxyx= sp.symbols('zxyx')
""""""
zxyy= sp.symbols('zxyy')
""""""
zxyz= sp.symbols('zxyz')
""""""
zxzx= sp.symbols('zxzx')
""""""
zxzy= sp.symbols('zxzy')
""""""
zxzz= sp.symbols('zxzz')
""""""
zyxx= sp.symbols('zyxx')
""""""
zyxy= sp.symbols('zyxy')
""""""
zyxz= sp.symbols('zyxz')
""""""
zyyx= sp.symbols('zyyx')
""""""
zyyy= sp.symbols('zyyy')
""""""
zyyz= sp.symbols('zyyz')
""""""
zyzx= sp.symbols('zyzx')
""""""
zyzy= sp.symbols('zyzy')
""""""
zyzz= sp.symbols('zyzz')
""""""
zzxx= sp.symbols('zzxx')
""""""
zzxy= sp.symbols('zzxy')
""""""
zzxz= sp.symbols('zzxz')
""""""
zzyx= sp.symbols('zzyx')
""""""
zzyy= sp.symbols('zzyy')
""""""
zzyz= sp.symbols('zzyz')
""""""
zzzx= sp.symbols('zzzx')
""""""
zzzy= sp.symbols('zzzy')
""""""
zzzz= sp.symbols('zzzz')
""""""

sin = sp.sin
cos = sp.cos
theta = sp.symbols('theta', real=True)
"""Angle of incidence

real = True
"""
phi = sp.symbols('phi', real=True)
"""Azimuthal angle

real = True
"""
psi = sp.symbols('psi', real=True)
"""Arbitrary phase shift

real = True
"""
Fx = sp.symbols('Fx')
""""""
Fy = sp.symbols('Fy')
""""""
Fz = sp.symbols('Fz')
""""""
I = 1j
conjugate = sp.conjugate



import unittest

import numpy as np
import shgpy
import shgpy.fformgen
import shgpy.tensor_definitions as td
import shgpy.shg_symbols as S
import logging
import random
from shgpy.plotter import easy_plot

_logger = logging.getLogger(__name__)

MANUAL = False


class TestfFormGen(unittest.TestCase):

    def test_generate_uncontracted_fourier_transforms(self):
        shgpy.fformgen.generate_uncontracted_fourier_transforms(0.1745, 'tests/fform/uft10deg')
        shgpy.fformgen.generate_uncontracted_fourier_transforms_symb('tests/fform/uft_theta')
    
    def test_generate_contracted_fourier_transforms(self):

        t_dip = shgpy.make_tensor_complex(shgpy.transform(shgpy.particularize(td.dipole['T_d']), shgpy.rotation_matrix_from_two_vectors(np.array([1, 1, 0]), np.array([0, 0, 1]))))
        t_quad = shgpy.make_tensor_complex(np.zeros(shape=(3,3,3,3), dtype=object))
        save_filename = 'tests/fform/T_d-S_2-S_2(110)-particularized-complex.p'
        shgpy.fformgen.generate_contracted_fourier_transforms_complex(save_filename, 'tests/fform/uft10deg', t_dip, t_quad, ndigits=4)
        fform = shgpy.load_fform(save_filename)
        fform.apply_phase_shift(S.psi)
        form = shgpy.fform_to_form(fform)
        dat = shgpy.form_to_dat(form, {k:random.uniform(-1, 1) for k in form.get_free_symbols() if k != S.phi}, 1000)
        if MANUAL:
            easy_plot([dat], [{'linestyle':'-', 'color':'blue'}], dat.get_keys())



import matplotlib.pyplot as plt
import numpy as np


def easy_plot(list_of_dats, list_of_param_dicts, pcs_to_include, show_plot=True, filename=None, show_legend=False):
    """An easy linear plotting routine for SHG data.

    Parameters
    ----------
    list_of_dats : list of DataContainer
        List of :class:`~shgpy.core.data_handler.DataContainer` instances to plot.
    list_of_param_dicts : list of dict
        List of `matplotlib` parameter dictionaries, each corresponding to
        a particular `DataContainer`.
    pcs_to_include : list of str
        List of polarization combinations to include in the plot.
    show_plot : bool
        Defaults to True
    filename : str or NoneType
        If `None`, no figure is saved. Otherwise, a figure is saved
        at the location specified by `filename`.
    show_legend : bool
        Defaults to `False`

    """
    n = len(pcs_to_include)
    if n > 4:
        raise ValueError('easy_plot only accepts a maximum of 4 polarization combos.')
    nrows = [1, 1, 1, 2][n-1]
    ncols = [1, 2, 3, 2][n-1]
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, sharey=True, squeeze=False)
    for pc, ax in zip(pcs_to_include, axs.flatten()):
        for dat, kwparams in zip(list_of_dats, list_of_param_dicts):
            xdata, ydata = dat.get_pc(pc, 'radians')
            ax.plot(xdata, ydata, **kwparams)
        ax.set_title(pc)
        if show_legend:
            ax.legend(loc='upper right')
    if filename is not None:
        plt.savefig(filename)
    if show_plot:
        plt.show()


def easy_polar_plot(list_of_dats, list_of_param_dicts, pcs_to_include, show_plot=True, filename=None, show_legend=False):
    """An easy polar plotting routine for SHG data.

    Parameters
    ----------
    list_of_dats : list of DataContainer
        List of :class:`~shgpy.core.data_handler.DataContainer` instances to plot.
    list_of_param_dicts : list of dict
        List of `matplotlib` parameter dictionaries, each corresponding to
        a particular `DataContainer`.
    pcs_to_include : list of str
        List of polarization combinations to include in the plot.
    show_plot : bool
        Defaults to True
    filename : str or NoneType
        If `None`, no figure is saved. Otherwise, a figure is saved
        at the location specified by `filename`.
    show_legend : bool
        Defaults to `False`

    """

    n = len(pcs_to_include)
    fig, axs = plt.subplots(1, n, subplot_kw=dict(projection='polar'), sharey=True)

    for pc, ax in zip(pcs_to_include, axs.flatten()):
        for dat, kwparams in zip(list_of_dats, list_of_param_dicts):
            xdata, ydata = dat.get_pc(pc, 'radians')
            ax.plot(xdata, ydata, **kwparams)
        ax.set_title(pc)
        if show_legend:
            ax.legend(loc='upper right')
    for pc,ax in zip(pcs_to_include, axs):
        ax.spines['polar'].set_visible(False)
        ax.set_yticklabels([])
        ax.set_xticks([0, np.pi/2, np.pi, 3*np.pi/2])
        ax.set_xticklabels(['', '', '', '', ''])
        ax.set_rticks([max([dat.get_maxval()[1] for dat in list_of_dats])])
        ax.set_rmin(0)
        ax.set_title(pc)
        ax.set_facecolor('#ffffff')
    if filename is not None:
        plt.savefig(filename)
    if show_plot:
        plt.show()



"""Module for (non-Fourier) formula derivation

This module provides a number of routines for calculating the SHG 
response of a given material (encoded by its susceptbility tensor,
see :class:`~shgpy.tensor_definitions` ). These routines are much
more straightforward than those found in :class:`~shgpy.fformgen`,
but are only suited for the most simple problems because the
conversion from formulas to Fourier formulas (which is all but
necessary for efficient fitting functionality) is typically quite
slow.

"""
import numpy as np
import sympy as sp
from .core import (
    FormContainer,
    tensor_contract,
    tensor_product,
)
from . import shg_symbols as S


def formgen_just_dipole_complex(t1, theta):
    """Generate formula assuming dipole SHG with complex coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    This routine differs from :func:`~shgpy.formgen.formgen_just_dipole_real` 
    only in the sense that the computed intensity function is computed by
    computing the modulus-squared (rather than ``**2``) of the polarization.
    For this reason, it is usually suggested to explicitly substitute 
    ``x -> real_x+1j*imag_x`` for each `x` in `t1`. See :func:`~shgpy.core.utilities.make_tensor_complex`
    and the tutorial.

    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    PP = Pp[0]*sp.conjugate(Pp[0])+Pp[2]*sp.conjugate(Pp[2])
    PS = Pp[1]*sp.conjugate(Pp[1])
    SP = Ps[0]*sp.conjugate(Ps[0])+Ps[2]*sp.conjugate(Ps[2])
    SS = Ps[1]*sp.conjugate(Ps[1])
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_just_dipole_real(t1, theta):
    """Generate formula assuming dipole SHG with real coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    This routine differs from :func:`~shgpy.formgen.formgen_just_dipole_complex` 
    only in the sense that the computed intensity function is computed by
    computing ``**2`` (rather than the modulus squared) of the polarization.
    This is usually easier as long as you're not worrying about the imaginary
    component of the susceptibility (valid away from resonance in the presence
    of time-reversal symmetry).
    

    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    PP = Pp[0]**2+Pp[2]**2
    PS = Pp[1]**2
    SP = Ps[0]**2+Ps[2]**2
    SS = Ps[1]**2
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_dipole_quadrupole_real(t1, t2, theta):
    """Generate formula assuming dipole+quadrupole SHG with real coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG dipole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    t2 : ndarray of sympy.Expr
        SHG quadrupole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    See note in :func:`~shgpy.formgen.formgen_just_dipole_real`.
    
    """
    c = sp.cos(theta)
    s = sp.sin(theta)
    kin = np.array([s, 0, c], dtype=object)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    rotated_qtensor = tensor_contract(tensor_product(R, R, R, R, t2), [[1, 8], [3, 9], [5, 10], [7, 11]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Qs = tensor_contract(tensor_product(rotated_qtensor, kin, Fs, Fs), [[1, 4], [2, 5], [3, 6]])
    Qp = tensor_contract(tensor_product(rotated_qtensor, kin, Fp, Fp), [[1, 4], [2, 5], [3, 6]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    Qs -= np.dot(kout, Qs)*kout
    Qp -= np.dot(kout, Qp)*kout
    PP = Pp[0]**2+Pp[2]**2+Qp[0]**2+Qp[2]**2
    PS = Pp[1]**2+Qp[1]**2
    SP = Ps[0]**2+Ps[2]**2+Qs[0]**2+Qs[2]**2
    SS = Ps[1]**2+Qs[1]**2
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})


def formgen_dipole_quadrupole_complex(t1, t2, theta):
    """Generate formula assuming dipole+quadrupole SHG with complex coefficients.

    Parameters
    ----------
    t1 : ndarray of sympy.Expr
        SHG dipole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    t2 : ndarray of sympy.Expr
        SHG quadrupole susceptibility tensor; see :class:`~shgpy.tensor_definitions`.
    theta : float or sympy.Symbol
        Angle of incidence

    Returns
    -------
    fform : FormContainer
        Instance of :class:`~shgpy.core.data_handler.FormContainer`.

    Notes
    -----
    See note in :func:`~shgpy.formgen.formgen_just_dipole_complex`.
    
    """
    theta = sp.sympify(theta)
    if not theta.is_real:
        raise ValueError('theta must be a real variable (did you forget real=True in sympy.symbols?)')
    c = sp.cos(theta)
    s = sp.sin(theta)
    kin = np.array([s, 0, c], dtype=object)
    kout = np.array([s, 0, -c], dtype=object)
    Fp = np.array([-c, 0, s], dtype=object)
    Fs = np.array([0, 1, 0], dtype=object)
    R = np.array([[sp.cos(S.phi), -sp.sin(S.phi), 0], [sp.sin(S.phi), sp.cos(S.phi), 0], [0, 0, 1]])
    rotated_tensor = tensor_contract(tensor_product(R, R, R, t1), [[1, 6], [3, 7], [5, 8]])
    rotated_qtensor = tensor_contract(tensor_product(R, R, R, R, t2), [[1, 8], [3, 9], [5, 10], [7, 11]])
    Ps = tensor_contract(tensor_product(rotated_tensor, Fs, Fs), [[1, 3], [2, 4]])
    Pp = tensor_contract(tensor_product(rotated_tensor, Fp, Fp), [[1, 3], [2, 4]])
    Qs = tensor_contract(tensor_product(rotated_qtensor, kin, Fs, Fs), [[1, 4], [2, 5], [3, 6]])
    Qp = tensor_contract(tensor_product(rotated_qtensor, kin, Fp, Fp), [[1, 4], [2, 5], [3, 6]])
    Ps -= np.dot(kout, Ps)*kout
    Pp -= np.dot(kout, Pp)*kout
    Qs -= np.dot(kout, Qs)*kout
    Qp -= np.dot(kout, Qp)*kout
    Sp = Pp+1j*Qp
    Ss = Ps+1j*Qs
    PP = Sp[0]*sp.conjugate(Sp[0])+Sp[2]*sp.conjugate(Sp[2])
    PS = Sp[1]*sp.conjugate(Sp[1])
    SP = Ss[0]*sp.conjugate(Ss[0])+Ss[2]*sp.conjugate(Ss[2])
    SS = Ss[1]*sp.conjugate(Ss[1])
    return FormContainer({'PP':PP, 'PS':PS, 'SP':SP, 'SS':SS})
